/**
Minidaemon ADT Implementation, provided by Mellanox, MPI Team
**/
#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/utsname.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#ifdef MAC_OSX
#include <sys/wait.h>
#else
#include <wait.h>
#endif
#include <assert.h>
#include <fcntl.h>
#include <time.h>

#include "minidaemon.h"

#define RSH_CMD	"/usr/bin/rsh"
#define SSH_CMD	"/usr/bin/ssh"
#define SSH_ARG "-q"

 
#define MAX_MESSAGE_SIZE 1024
#define MINIDAEMON_PORT 4000
#define ACCEPT_TIMEOUT 25
#define NO_TIMEOUT 1000

#define MIN(a,b) ((a) < (b) ? (a) : (b))



//#define CH_LIST_FOREACH_MD_NODES 


//typedef enum {MD_INIT_DATA = 0, QP_LID_INFO , ALL_QP_LID_INFO, STATUS_REQUEST, MD_STATUS , KILL} MPI_msg;
//typedef enum {MD_INIT=0, MD_PROC, MD_NORM_TERM, MD_FAIL} MD_status;

typedef enum {CH_JOB_EXT = -1,P_OK = 0, P_KILLED, STATUS_REQUEST, STATUS , KILL} MD_message;
typedef enum {MD_INIT = 0, MD_QP_LID, MD_WORK, MD_TERM } MD_phase;

typedef enum {MD_NORMAL, MD_UNKNOWN} MD_status;

/*Timeout values for WORK stage */
static const int send2father_work_timeout = 10;
static const int recv_parent_work_timeout = 20;
static const int recv_children_work_timeout = 20;
static const int recv_jobs_work_timeout = NO_TIMEOUT;

/* Timeout values for TERM stage */
static const int send2father_term_timeout = NO_TIMEOUT;
static const int recv_parent_term_timeout = NO_TIMEOUT;
static const int recv_children_term_timeout = 20;
static const int recv_jobs_term_timeout = 10;

/** These 2 structures are for temporarily testing only **/
/**-----------------------------------------------------------**/

struct minidaemon_t {
	//private data members
	const char	*parent_name;
	int			tree_width; // md_num
	int			ch_num; // total process number, i.e the length of children_md_arr
	
	int			*status_arr; //these 2 members are for status maintance
	int			status_arr_len;
	
	MD_phase	phase; // phase of execution
	
	MD_entry	children_md_arr; // The list of children and parent minidaemon process and own jobs
	
	ChildrenList childrenList; // The list of parameters for minidaemon itself and its subtree
	
	//public methods //TODO this methods would be implemented in xlauncher ver 2.0
	/*void 		(*timeout_handler)();
	void 		(*data_handler)();*/
	int			curr_running_jobs;
	int			curr_running_mds;
	
	//timeout variables
	//TODO init it in ...init();
	int 		send2father_timeout ;
	int 		recv_parent_timeout ;
	int 		recv_children_timeout ;
	int 		recv_jobs_timeout ;
	
	const char*	mpi_command_params;
	const char* mpi_command;
	int			mpi_param_len;
	
	int 		max_fd_num;
	
	int			isParent;
	
};
static struct minidaemon_t md_entity;

struct md_entry_t {
	int fd;
	MD_status stat;
	struct sockaddr * sock_addr;
}; //TODO not pointer

struct childrenList_t {
    char *hostname;
    char *device;
    int pid; //TODO convert it to status
    int port;
	int rank;
    /*int control_socket;
    process_state state;*/ //TODO
};

int resp_status_def = 0;

typedef int ChildrenListIterator;

//TODO all functions that use 'private' data member Minidaemon should be static
static int 					children_list_comp (const void *p1, const void *p2);
static ChildrenListIterator 	getListNextNode(ChildrenListIterator it, int offset);

static void 	minidaemon_socket_create();
void 			minidaemon_connect_to_parent();
void 			md_listen_parent();
void 			md_read_int(int * buff, int sd);
void 			md_read_string(char * buff, int sd);
void 			md_start_node_tree();
void 			md_start_node(ChildrenListIterator i);
void 			run_own_jobs();
//void 			minidaemonListen();
void 			minidaemonGeneralListen();
void 			timeout_handler (int delta_time);
void 			children_proc_handler(int ind,int time_delta);
void 			md_handler(int sd);
void 			sendStatusMessage(int sd);
void 			rebuildMask(fd_set * mask);
int 			updateSocketStatus(int ind,int time_delta);
void 			startTerminationProcess();
void 			forced_cleanup_handler();
void 			req_status_handler(const char * buff);
void 			sendInitMessage();
void 			status_msg_handler();
void 			sendTerminationMessage(int sd);
void 			md_base_init(int ch_num, int width, const char * command_params,const char * command);


static int children_list_comp (const void *p1, const void *p2) {
	ChildrenList proc1 = (ChildrenList ) p1;
	ChildrenList proc2 = (ChildrenList ) p2;
	return strcmp(proc1->hostname, proc2->hostname);
}

static ChildrenListIterator getListNextNode(ChildrenListIterator it, int offset) {
	ChildrenListIterator it_next = it + 1;
	/*if ((it == (md_entity.ch_num -1)) && ()) 
		return md_entity.ch_num;*/
	if (it < 0 || offset < 0) {
		MD_USR_ERROR("<getListNextNode> Bad Iterator received");
	}
	
	if (it_next >= md_entity.ch_num )
		return md_entity.ch_num;
		
	while (offset > 0 && it_next < md_entity.ch_num) {
		while (children_list_comp( &md_entity.childrenList[it], &md_entity.childrenList[it_next])) {
			if ( (++it_next >= md_entity.ch_num) )
				return md_entity.ch_num; // return there maximum iterator value +1 in order to notify list end
		}
		--offset;
		it=it_next++;
	}
	return it_next; //the actual index or the end of the list (ch_num) should be returned
}

/**
Create a root instance of 
This function should be
**/
void minidaemon_create(process * procList, int nproc, int width, const char * command_params, const char * command) {
	int i;
	md_base_init(nproc, width, command_params, command);
	for (i = 0; i < nproc; ++i) {
		//(process) (childrenList[i]) = procList[i]; //TODO
		md_entity.childrenList[i] = *((ChildrenList)&(procList[i]));
		md_entity.childrenList[i].rank = i;
	}
	
	qsort(md_entity.childrenList, nproc, sizeof(ChildrenList), children_list_comp);
	
	md_entity.isParent	 = 1;
}

// Init all data member structures
void minidaemon_init ( const char * par_name , int ch_num, int width, const char * command_params,const char * command) {

	md_base_init(ch_num, width, command_params, command);
	md_entity.parent_name 		= par_name;
	md_entity.isParent = 0;
	//TODO other initialization
}

void md_base_init(int ch_num, int width, const char * command_params,const char * command) {
	if ( !(md_entity.children_md_arr = malloc(width * sizeof(MD_entry))) ||
	     !(md_entity.status_arr = malloc(ch_num *sizeof(char))) ||
		 !(md_entity.childrenList = malloc(ch_num * sizeof(ChildrenList))) ) {
		MD_SYS_ERROR("<minidaemon_init> malloc failed");
	} 

	md_entity.ch_num 			= ch_num;
	md_entity.tree_width 		= width;
	md_entity.status_arr_len 	= 2; // status array should contain all
	
	md_entity.mpi_command_params = command_params;
	md_entity.mpi_param_len		 = strlen(command_params);
	md_entity.mpi_command 		 = command;
	
	md_entity.max_fd_num = 0;
}

//create parent socket and children minidaemon sockets
static void minidaemon_socket_create() {
	
	struct hostent* host;
	struct utsname name;
	struct sockaddr_in srv_sock_name,clnt_sock_name;
	int s;
	long arg;
	
	if ((s = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
		perror ("socket");
		exit(1);
	}
	
	
	memset ((char *)&srv_sock_name, '\0', sizeof(srv_sock_name));
	srv_sock_name.sin_port = htons (MINIDAEMON_PORT);
	srv_sock_name.sin_family = AF_INET;
	
	/* determine host system name and internet address*/
	if (uname(&name) == -1) {
		perror ("uname");
		exit(1);
	}
	
	if ( (host=gethostbyname(name.nodename)) == NULL) {
		perror ("gethostbyname");
		exit(1);
	} 
	
	memcpy ( (char*) &srv_sock_name.sin_addr, host->h_addr_list, host->h_length);
	
	if (bind (s, (struct sockaddr *) &srv_sock_name, sizeof (srv_sock_name)) < 0) {
		perror ("bind");
		close (s);
		exit(1);
	}
	
		// Set non-blocking 
	if( (arg = fcntl(s, F_GETFL, NULL)) == 0) { 
		fprintf(stderr, "Error fcntl(..., F_GETFL) (%s)\n", strerror(errno)); 
		exit(0); 
	} 
	arg |= O_NONBLOCK; 
	if( fcntl(s, F_SETFL, arg) == 0) { 
		fprintf(stderr, "Error fcntl(..., F_SETFL) (%s)\n", strerror(errno)); 
		exit(0); 
	} 
	
	//prepare to listen cp for multiple connections	
	if (listen (s, md_entity.tree_width) ==-1 ) {
		perror ("listen");
		close (s);
		exit(1);
	}
	
	int loop = md_entity.tree_width;
	//TODO: spinoff this block
	struct timeval timeout;
	timeout.tv_sec = ACCEPT_TIMEOUT;
	timeout.tv_usec = 0;
	int res;
	fd_set mask;
	int client_len = sizeof(clnt_sock_name);
	while (loop) {
		switch (res = select(s+1, &mask, NULL, NULL, &timeout)) {
			case -1:
				/*signal or error were received*/
				/*perror ("select");
				close (s);
				exit(1);*/
			
			case 0 :
				/* timeout was reached */
				fprintf(stderr,"Timeout problem when connecting to other minidaemons");
				exit(1);
			default:
				if (res < 0) {
					perror ("Bad select result\n"); //TODO: remove this line
					exit(1);
				}
				if (FD_ISSET (s, &mask)) { //TODO assert
					//TODO : yet another try
					/*	There may not always be a connection waiting after a SIGIO is delivered or select(2) or poll(2)
					 return a readability event because the  connection might  have  been  removed  by  an asynchronous
					 network error or another thread before accept is called.  If this happens then the call will block
					 waiting for the next connection to arrive.  To ensure that accept never blocks, the passed socket 
					 s needs to have the  O_NONBLOCK  flag  set
					*/
					if ( (md_entity.children_md_arr[loop].fd = 
						accept(s, (struct sockaddr *) &clnt_sock_name,&client_len)) == -1 ) {
						
						perror ("listen");
						close (s);
						exit(1);
					}
					md_entity.max_fd_num =
					 ( md_entity.max_fd_num > md_entity.children_md_arr[loop].fd 
					 ? md_entity.max_fd_num : md_entity.children_md_arr[loop].fd) ;
					 
					sendInitMessage(md_entity.children_md_arr[loop].fd);
					//children_md_arr[loop].sock_addr = clnt_sock_name; //TODO ?
					--loop;
				}
				break;
		}
	}
	return;	
}

// Start listening to messagesvoid req_status_handler(const char * buff) from other Minidaemons
void minidaemon_run () {

	ChildrenListIterator i;
	int node_number = 0;
	if (!md_entity.isParent) {
		minidaemon_connect_to_parent();
	}
	md_start_node_tree();
	run_own_jobs();
	minidaemon_socket_create();
	
	/* Here are 4 different stages :
	1. - INIT STATE : Get parameters from parent minidaemon (this stage should be done in minidaemon_connect_to_parent()
	2. - QP_LID STAGE : Listen to and handle QP_LID messages in order to run own jobs and pass params to children minidaemons
	3. - WORK STAGE : Listen to STATUS messages from other MD's and keep I-AM-ALIVE on
	4. - TERMINATION STAGE : Listen to messages in order to verify clear shutdown
	*/
	
	minidaemonGeneralListen();
	exit(0);
}

void minidaemon_connect_to_parent() {
	struct hostent* host;
	struct utsname name;
	struct sockaddr_in srv_sock_name,clnt_sock_name;
	int s;
	long arg;
	int len;
	int i = 0;
	
	if ((s = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
		perror ("socket");
		exit(1);
	}
	
	memset ((char *)&srv_sock_name, '\0', sizeof(srv_sock_name));
	srv_sock_name.sin_port = htons (MINIDAEMON_PORT);
	srv_sock_name.sin_family = AF_INET;
	
	if ( (host=gethostbyname(md_entity.parent_name)) == NULL) {
		perror ("gethostbyname");
		exit(1);
	} 
	
	memcpy ( (char*) &srv_sock_name.sin_addr, host->h_addr_list, host->h_length);
	
	if (connect (s, (struct sockaddr *) &srv_sock_name, sizeof (srv_sock_name)) < 0) {
		perror ("connect");
		close (s);
		exit(1);
	}
	md_entity.max_fd_num = ( md_entity.max_fd_num > s ? md_entity.max_fd_num : s) ;
	/* 
	 Here we not need the communication to be non-blocking 
	 Minidaemon should receive data that is essential for future flow.
	 That is, without information about own processes and children minidaemon it can't continue
	*/
	md_listen_parent(s);
}

void md_listen_parent(int sd) {
	int i;
	while (i < md_entity.ch_num ) {
		md_read_string(md_entity.childrenList[i].hostname,sd);
		md_read_string(md_entity.childrenList[i].device,sd);
		md_read_int(&md_entity.childrenList[i].port,sd);
		md_read_int(&md_entity.childrenList[i].rank,sd);
		++i;
	}
}

void md_read_int(int * buff, int sd) {
	//TODO while instead of if
	if ( (read(sd,buff, sizeof(int)) < 0) || (*buff < 0)) {
		MD_USR_ERROR("Unexpected data : exiting");
	}
}

void md_read_string(char * buff,int sd) {
	//TODO while instead of if
	int len , n;
	if ( (n = read(sd,&len, sizeof(int)) < 0) || (len < 1)) {
		MD_USR_ERROR("Unexpected data : exiting");
	}
	n = 0;
	while ( n < len ) { //TODO possible catch of signal or check if n always > 0
		n += read(sd,buff+n,len-n);
	}
}



void md_start_node_tree() {
	// We have to pass through the first r nodes (own jobs)
	// Then we have to run execv command for each
	ChildrenListIterator i=0;
	while (i < md_entity.ch_num) {
		// here we passed through all the chidren of the "width+1" md nodes 
		md_start_node(i=getListNextNode(i,md_entity.tree_width));
		++md_entity.curr_running_mds;
	}
}

//TODO !!!! to send to each child minidaemon its exact list of processes
// minidaemon number p => list.get_node_by_iterator(p):(p+width)
void md_start_node(ChildrenListIterator i) {

	static const char  minidaemon_command[] = "/hpc/work/xalex/mvapich-0.9.7-release-659-before-changes/mpid/ch_gen2/process/minidaemon_client.exe ";
	
	if (i >= md_entity.ch_num )
		return;
	
	if ((md_entity.childrenList[i].pid=fork()) == 0) { //child process
		execl(SSH_CMD, SSH_CMD, SSH_ARG, md_entity.childrenList[i].hostname,
                minidaemon_command, md_entity.childrenList[0].hostname,md_entity.tree_width, NULL);
	// If we've reached this line, ssh failed
		MD_SYS_ERROR ("RSH/SSH command failed!");
	}
}

void run_own_jobs() {
	
	ChildrenListIterator i=0;
	int pd[2];
	const int add_param_len = 45; //the lengths of VIADEV_DEVICE etc.
	const int max_rank_len = 10; //the maximum length of rank number in symbols , i.e. 1000000 mpi jobs
	const int max_port_len = 8;
	char * curr_mpi_command_params = NULL;
	char rank_str[max_rank_len];
	char port_str[max_port_len];
	int curr_param_len = 0;
	
	do {
		/*itoa(md_entity.childrenList[i].rank,rank_str);
		itoa(md_entity.childrenList[i].rank,port_str);*/
		
		sprintf(rank_str,"%d\0",md_entity.childrenList[i].rank);
		sprintf(port_str,"%d\0",md_entity.childrenList[i].rank);
		
		curr_param_len = sizeof(char) *( md_entity.mpi_param_len + add_param_len
			+ strlen(md_entity.childrenList[i].device) + strlen(port_str) + strlen(rank_str));
		if (!curr_mpi_command_params) {
			if ((curr_mpi_command_params = malloc(curr_param_len)) == 0) {
				MD_SYS_ERROR("<run_own_jobs> : malloc failed\n");
			}
		} else {
			if (strlen(curr_mpi_command_params) < curr_param_len) {
				realloc(curr_mpi_command_params,curr_param_len);
			}
		}
		sprintf(curr_mpi_command_params, "%s VIADEV_DEVICE=%s VIADEV_PORT=%s  MPIRUN_RANK=%s",md_entity.mpi_command_params,
			md_entity.childrenList[i].device, md_entity.childrenList[i].port, rank_str);
			
		// create pipe in order to communicate with children process
		if (pipe(pd) == -1) {
			MD_SYS_ERROR("pipe");
		}
		md_entity.children_md_arr[i].fd = pd[0];
		if ((md_entity.childrenList[i].pid = fork())== 0) {  /* Affinity should be set in vianit.c */
			close(pd[0]); //closing read side of pipe in the child process
			execl(md_entity.mpi_command, curr_mpi_command_params); //pd[1] will be closed in the end of execution
			MD_SYS_ERROR("execl failed\n");
		}
		else {
			//close write side
			close(pd[1]);
			++i;
			++md_entity.curr_running_jobs;
		}
	} while ( i < md_entity.ch_num && children_list_comp(&md_entity.childrenList[0],&md_entity.childrenList[i])) ;
	
	free(curr_mpi_command_params);
	curr_mpi_command_params = NULL;
	//TODO while (i < MIN (md_entity.ch_num , getNextListNode(0,1))
}

//TODO rebuild minidaemonListen() to be a "Class" with general methods, passed by parameters
/*
void minidaemonListen() {

	
	int timeout = send2father_timeout;
	int status,res,stat;
	time_t time_wasted;
	char msg_buf[MAX_MESSAGE_SIZE];
	fd_set wr_mask; //TODO rd , not wr
	while (1) {
		rebuildMask(&wr_mask);
		switch (res = select(max_fd_num+1,&wr_mask,NULL,NULL,timeout)) {
			case -1:
				//Signal or error was received
			case 0:
				md_entity.timeout_handler();
				// timeout had happened
				if ((timeout = global_timeout - (time(NULL)-time_wasted)) > 1) {
					sendRequestMessage();
				} //TODO : delta instead of 1
				else {
					minidaemon_timeout_handler();
				}
				break;
			default:
				if (res < 0) { 
					perror ("Bad select result\n"); //TODO: remove this line
					exit(1);
				}
				int i;
				port;// Find all file descriptors that had been updated
				for (i = 0 ; i < children_md_num + md_entity.curr_running_jobs; ++i) {
					if (FD_ISSET (md_entity.children_md_arr[i].sd, &wr_mask)) {
						// Child process was ended
					/
						if (i < md_entity.curr_running_jobs) {
							stat = waitpid (cmd_entity.children_md_arr[i].pid,&status,WNOHANG);
							//TODO prevent race ! (i.e. sleep(2)
							status_buf[i]=stat;
							startTerminationProcess(i,stat);
							break; //TODO return instead of break
							//TODO wait for all children procs
						}
				
						minidaemonGeneralListen();// Child Minidaemon has sent us a message
						else { 
							//TODO all the sockets should be marked as non-blocking
						minidaemonGeneralListen();	//nsd = accept(children_md_arr[i].sd, &md_client, &md_client_len); -TODO-init stage only
							n = read(md_entity.children_md_arr[i].sd,buff,MAX_MESSAGE_SIZE);
							minidaemon_message_handler(buff,n);
							stat = updateSocketStatus(i);
							
						}
						break;
					}
				} // for
				timeout = (stat ? send2father_timeout : int(time() - time_wasted));
				break;
		} // switch
	} // while
	
}
*/
void minidaemonGeneralListen() {
	// Registrate Event Handlers, if any
	// listen to all fd'minidaemonGeneralListen()s, including self-pipes, if any

	md_entity.send2father_timeout 	= send2father_work_timeout;
	md_entity.recv_parent_timeout 	= recv_parent_work_timeout;
	md_entity.recv_children_timeout = recv_children_work_timeout;
	md_entity.recv_jobs_timeout 	= recv_jobs_work_timeout;
	
	int status,res,stat;
	time_t time_wasted;
	char msg_buf[MAX_MESSAGE_SIZE];
	fd_set wr_mask;
	struct timeval timeout;
	timeout.tv_usec = 0;
	fd_set fd_array;
	FD_ZERO(&fd_array);
	while (1) {
		// Wait for father and children to respond, no more than fixed timeout
		time_wasted = time(NULL);
		rebuildMask(&wr_mask); // should be also general function
		if (md_entity.phase == MD_WORK) {
			timeout.tv_sec = MIN(md_entity.send2father_timeout,md_entity.recv_children_timeout);
		}
		else {
			timeout.tv_sec = MIN(md_entity.recv_jobs_timeout,md_entity.recv_children_timeout);
		}
		
		switch (res = select(md_entity.max_fd_num+1,&fd_array,NULL,NULL,&timeout)) {
			case -1:
				//Signal or error was received
				perror("<minidaemonGeneralListen()> Unhandled signal\n");
				exit(1);
			case 0:
				//Is parent timeout ? 
				timeout_handler((int)(time(NULL) - time_wasted));
				break;
			default:
				// wait for confirmation status from children
				assert(res > 0);
				int i;
				// Find all file descriptors that had been updated
				for (i = 0 ; i < md_entity.curr_running_mds + md_entity.curr_running_jobs+ 1; ++i) {
					if (FD_ISSET(md_entity.children_md_arr[i].fd, &fd_array)) {
						// Child process was ended
						/* TODO : spin this block out as a separate function */
						if (i < md_entity.curr_running_jobs) {
							//TODO set appropriate timeout
							//TODO handle different cases
							children_proc_handler(i,(int)(time(NULL) - time_wasted));
						}
						//Parent or children minidaemons had sent us a message
						else {
							md_handler(md_entity.children_md_arr[i].fd); // for parent and children md messages
							updateSocketStatus(i,(int)(time(NULL) - time_wasted)); 
						}	
						//break;
					}
				} // for
				break;
		} // switch
	} // while
}
	
void timeout_handler (int delta_time) {

	// in this case, timeout_ch is a current "receive-from-children-or-parent" timeout
	// and timeout_other is timeout of sending to parent
	switch(md_entity.phase) {
	case MD_WORK :
		if (md_entity.send2father_timeout < md_entity.recv_children_timeout) {
			sendStatusMessage(md_entity.children_md_arr[md_entity.curr_running_jobs].fd);
			md_entity.send2father_timeout 	= send2father_work_timeout;
			md_entity.recv_children_timeout -= delta_time;
			assert(md_entity.recv_children_timeout > 0);
		}
		//We got timeout from children and/or parent
		else {
			startTerminationProcess(); 
			md_entity.recv_children_timeout = recv_children_term_timeout;
			md_entity.recv_jobs_timeout 	= recv_jobs_term_timeout;
		}
		break;
		
	// in this case, timeout_ch is a current "receive-from-chilren-or-parent" timeout
	// and timeout_other is timeout for own-jobs-termination	
	case MD_TERM :
		// here we should wait only for children process and children minidaemons	
		//TODO propagation delay, i.e. to_value = f(base_to_value,tree_depth)
		if (md_entity.curr_running_jobs !=0) {
			forced_cleanup_handler();
		}
		//the last status Message we've send
		sendStatusMessage(md_entity.children_md_arr[md_entity.curr_running_jobs].fd);
		exit(0);
		
	default:
		break;
	}
}

void children_proc_handler(int ind, int time_delta) {

	static int counter = -1 ;
	if (counter == -1 )
		counter = md_entity.curr_running_jobs ;
	
	//TODO
	int stat = waitpid (md_entity.childrenList[ind].pid,&md_entity.status_arr[ind],WNOHANG);
	if (counter == md_entity.curr_running_jobs) {
		md_entity.phase = MD_TERM;
		md_entity.recv_jobs_timeout = recv_jobs_term_timeout;
	}
	if ( counter == 0) {
		md_entity.recv_jobs_timeout = NO_TIMEOUT;
	}
	else {
		md_entity.recv_jobs_timeout-= time_delta;
	}
	--counter;
}


//TODO array of pointer to appropriate function , i.e. func_array[msg_buff[0]].handler();
void md_handler(int sd) {
	static int msg_buf[MAX_MESSAGE_SIZE]; //TODO spin it out
	int n = read(sd, msg_buf, MAX_MESSAGE_SIZE);
	if ( (n < 0) || (msg_buf == NULL) ) {
		perror ("read was failed in spite of select");
		exit(1); //TODO
	}
	switch ( msg_buf[0]) {
		
		case STATUS_REQUEST:
			sendStatusMessage(sd);
			break;
		case STATUS:
			status_msg_handler(msg_buf);
			break;
		case KILL:
			startTerminationProcess();
			break;
		default:
			perror("<minidaemon_message_handler> : Bad opcode received\n");
	}
	return;
	
}

//TODO pass index
void status_msg_handler(const int * msg_buf) {

	if (md_entity.phase == MD_TERM || (msg_buf[1] == MD_TERM))  {
		int i;
		md_entity.status_arr_len = md_entity.ch_num;
		memcpy(&md_entity.status_arr[getListNextNode(0,i)],msg_buf, sizeof(int*));
		for (i = 1; i < md_entity.status_arr_len; ++i) {
			md_entity.status_arr[i] = MD_UNKNOWN;
		}
		md_entity.status_arr_len = md_entity.ch_num;
		startTerminationProcess();
	} 
}

void sendStatusMessage(int sd) {
	write(sd,md_entity.status_arr,md_entity.status_arr_len);
}

// TODO : define which set - parent, md or proc children should be set
void rebuildMask(fd_set * mask) {
	int i;
	FD_ZERO(mask);
	//TODO define which set - parent, md or proc children should be set
	for (i = 0 ; i <  md_entity.curr_running_mds + md_entity.curr_running_jobs; ++i) {
		FD_SET (md_entity.children_md_arr[i].fd, mask) ;
	}
}


int updateSocketStatus(int i,int time_delta) {

	int res;
	static int num_of_md_responded = 0;
	if (md_entity.children_md_arr[i].stat != resp_status_def) {
		md_entity.children_md_arr[i].stat = resp_status_def;
		++num_of_md_responded;
	
		if (res=(num_of_md_responded == md_entity.curr_running_mds))  {
			num_of_md_responded = 0;
			resp_status_def = !resp_status_def;
			md_entity.recv_children_timeout = recv_children_work_timeout;
		}
		else if (num_of_md_responded == 0) {
			md_entity.recv_children_timeout-=time_delta;
		}
	}
	return res;
}

void startTerminationProcess() {
	int i ;
	if (md_entity.phase != MD_TERM) {
		md_entity.phase = MD_TERM;
		md_entity.status_arr[0] = MD_TERM;
		for (i = 0 ; i < md_entity.curr_running_mds+1; ++i) {
			// send status message to other MD (with termination status)
			//TODO check borders of i
			sendTerminationMessage(md_entity.children_md_arr[md_entity.curr_running_jobs+i].fd);
		}
	}
}

void sendTerminationMessage(int sd) {
	if (write (sd,md_entity.status_arr,md_entity.status_arr_len*sizeof(int)) < 0)
		MD_SYS_ERROR("write");
}

void forced_cleanup_handler() {
	//send kill to its own jobs
	ChildrenListIterator i;
	//send forced kill (kill -9) to its own jobs
	for (i=0; i < getListNextNode(i,1) ; ++i) {
		kill(md_entity.childrenList[i].pid,SIGKILL);
	}
	for (i=0; i < getListNextNode(i,1) ; ++i) {
		waitpid(md_entity.childrenList[i].pid,&md_entity.status_arr[i],WNOHANG);
	}
	// send status message to other MD (with termination status)
}



void req_status_handler(const char * buff) {
	switch (buff[1]) {
	case MD_INIT:
		sendInitMessage();
	default:
		sendStatusMessage(md_entity.children_md_arr[md_entity.curr_running_mds].fd);
	}
}

//TODO receive width times at the client side !!!
void sendInitMessage() {
	ChildrenListIterator it = getListNextNode(0,1);
	int node_counter = 0, len;
	while (it < md_entity.ch_num ) {
	
		len=strlen(md_entity.childrenList[it].hostname)*sizeof(char);
		write(md_entity.children_md_arr[md_entity.curr_running_jobs + 1 + node_counter].fd,
			&len,sizeof(int));
		
		write(md_entity.children_md_arr[md_entity.curr_running_jobs + 1 + node_counter].fd,
			md_entity.childrenList[it].hostname,len);
		
		len=strlen(md_entity.childrenList[it].device)*sizeof(char);	
		write(md_entity.children_md_arr[md_entity.curr_running_jobs + 1 + node_counter].fd,
			&len,sizeof(int));
		
		write(md_entity.children_md_arr[md_entity.curr_running_jobs + 1 + node_counter].fd,
			md_entity.childrenList[it].device,len);
		
		write(md_entity.children_md_arr[md_entity.curr_running_jobs + 1 + node_counter].fd,
			&md_entity.childrenList[it].port,sizeof(int));
		
		write(md_entity.children_md_arr[md_entity.curr_running_jobs + 1 + node_counter].fd,
			&md_entity.childrenList[it].rank,sizeof(int));
		
		if (++it < md_entity.ch_num && !children_list_comp( &md_entity.childrenList[it-1], &md_entity.childrenList[it])) {
			++node_counter;
		}
	}
}
